import { Signal } from "./../SideralObject";


/**
 * Signals emitted by a SideralObject
 */
export interface ISignals {

    /**
     * Signal emitted everytime the object is updating
     */
    update: Signal;

    /**
     * Signal emitted when a property (which the name is passed in parameter) has changed
     */
    propChange: Signal;

    /**
     * Signal emitted when a new SideralObject is adding to the current object
     */
    addChild: Signal;

    /**
     * Signal emitted when a SideralObject is removing from the current object
     */
    removeChild: Signal;
}


/**
 * Signals emitted by the Keyboard Object
 */
export interface IKeyboardSignals extends ISignals {

    /**
     * Signal emitted everytime a key is pressed or released
     */
    keyChange: Signal;

    /**
     * Signal emitted when a key is pressed
     */
    keyPress: Signal;

    /**
     * Signal emitted when a key is released
     */
    keyRelease: Signal;
}


/**
 * Signals emitted by a Module
 */
export interface IModuleSignals extends ISignals {

    /**
     * Signal emitted on click on the module
     */
    click: Signal;

    /**
     * Signal emitted on doubleclick on the module
     */
    doubleClick: Signal;

    /**
     * Signal emitted on right click on the module
     */
    rightClick: Signal;

    /**
     * When mouse is hover
     */
    hover: Signal;

    /**
     * When mouse enter in hover
     */
    hoverStart: Signal;

    /**
     * When mouse is not hover anymore
     */
    hoverEnd: Signal;
}

/**
 * Signals emitted by the Scene object
 */
export interface ISceneSignals extends IModuleSignals {
    /**
     * Signal emitted during the loading of assets (used only if the scene is considered has the loader of the game)
     */
    progress: Signal;
}

/**
 * Signals emitted by a Skill
 */
export interface ISkillSignals extends ISignals {

    /**
     * When the preparation timer is started
     */
    preparationStart: Signal;

    /**
     * When the preparation timer is updating
     */
    preparationUpdate: Signal;

    /**
     * When the preparation timer is completed
     */
    preparationComplete: Signal;

    /**
     * When the skill is started
     */
    skillStart: Signal;

    /**
     * When the skill is updating
     */
    skillUpdate: Signal;

    /**
     * When the skill is completed
     */
    skillComplete: Signal;

    /**
     * When the reload timer is started
     */
    reloadStart: Signal;

    /**
     * When the reload timer is updating
     */
    reloadUpdate: Signal;

    /**
     * When the reload timer is completed
     */
    reloadComplete: Signal;
}

/**
 * Signals of Physic
 */
export interface IEntitySignal extends IModuleSignals {

    /**
     * When entity is entering in collision with an other entity
     */
    beginCollision: Signal;

    /**
     * Signal emitted every update if the entity is currently in collision with an other entity
     */
    collision: Signal;

    /**
     * Signal emitted every time the entity hit a wall
     */
    wallCollision: Signal;

    /**
     * When entity is not in collision anymore
     */
    endCollision: Signal;
}

/**
 * Signal for graphic
 */
export interface IGraphicSignal extends IModuleSignals {

    /**
     * When the state of the graphic has changed
     */
    stateChange: Signal;
}
