import { SideralObject } from "./../SideralObject";

/**
 * This class is used to generate events and properly destroy it when a SideralObject is killed 
 */
export class Signal {

    /* ATTRIBUTES */

    /**
     * The parent of the Signal
     * @readonly
     */
    readonly parent: SideralObject;

    /**
     * List of listeners connected to the signal
     * @private
     */
    _listeners: Array<Function> = [];

    /**
     * Function triggered when a listener is bound to the signal
     */
    onBind: Function;

    /**
     * Function triggered when a listener is unbound to the signal
     */
    onUnbind: Function;


    /* METHODS */

    /**
     * @constructor
     * @param parent - The parent of the signal
     */
    constructor (parent: SideralObject, onBind?: Function, onUnbind?: Function) {
        if (!(parent instanceof SideralObject)) {
            throw new Error("Signal.constructor: an instance of SideralObject must be passed on the constructor (the owner of the signal).");
        }

        this.parent     = parent;
        this.onBind     = onBind;
        this.onUnbind   = onUnbind;
    }

    /**
     * Dispatch a signal to all listeners with parameters
     * @param params - Parameters
     */
    dispatch (...params: Array<any>): void {
        if (typeof this._listeners === "undefined") {
            console.log([this, this.parent], this._listeners);
        }

        this._listeners.forEach(listener => listener(...params));
    }

    /**
     * Add a listener to the current signal
     * @param listener - The listener
     * @returns The listener
     */
    add (listener: Function): Function {
        this._listeners.push(listener);

        if (this.onBind) {
            this.onBind(listener);
        }

        return listener;
    }

    /**
     * Remove a listener from the signal
     * @param listener - The listener to remove
     * @returns True if listener has been removed
     */
    remove (listener: Function): boolean {
        const nextListeners = this._listeners.filter(x => x !== listener),
            hasRemoved      = nextListeners.length !== this._listeners.length;

        this._listeners = nextListeners;

        if (this.onUnbind && hasRemoved) {
            this.onUnbind(listener);
        }

        return hasRemoved;
    }

    /**
     * Remove all listeners from the signal
     */
    removeAll (): void {
        this._listeners.forEach(this.remove.bind(this));
        this._listeners = [];
    }

    /**
     * Bind a listener by the first parameter of a trigger
     * @param param - The parameter binded to the listener
     * @param listener - The listener
     * @returns The listener
     */
    bind (param: string | Array<string>, listener: Function): Function {
        const params: Array<string> = [].concat(param);

        return this.add(function (key: any) {
            if (typeof key === "string" && params.find(x => x === key)) {
                listener(...[].slice.call(arguments, 1));
            }
        });
    }
}